/*****************************************************************************
 *
 * FILE:	mod_session2.c
 * DESCRIPTION:	Wyvern: session management module
 * DATE:	Tue, Jul  6 2004
 * UPDATED:	Tue, Nov  9 2004
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2004 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2004 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_session2.c,v 1.2 2004/11/09 14:56:39 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif	/* HAVE_SYS_TIME_H */
#include <sqlite.h>
#include <wyvern.h>
#include "session.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	SESSION_CONF		WYVERN_CONF_DIR "/session2.conf"

#define WM_SID_INVALID	(-1)
#define WM_SID_VALID	(0)
#define WM_SID_EXPIRED	(1)

#define	HTTP_PORT	(80)

typedef struct _wmSession {
  const char *		name;	/* server name */
  const char *		path;	/* region for session use */
  size_t		plen;	/* path length */
  const char *		page;	/* path for login (top) page */
  struct _wmSession *	next;
} wmSession;

/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void		wm_session_init(void);
static void		wm_session_finish(void);
static void		wm_session_register_hooks(void);
static int		wm_validate_session(Request *);
static void		parse_config(sxml_node_t *);
static void		parse_error(const char *);
static wmSession *	lookup(const char *, const char *);
static int		validate_session_id(const char *, long *, long *);
static char *		get_sid_from_query(char *);
static char *		get_sid_from_cookie(char *);
static int		parse_string(char *, char **, char **);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static wmSession *	smgr = NULL;
static sqlite *		sdb  = NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wm_session_init
 * Description:		initialize CGI object
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_session_init(void)
{
  register FILE *	fin;

  fin = fopen(SESSION_CONF, "r");
  if (fin != NULL) {
    sxml_node_t *	root;

    root = sxml_parse_file(fileno(fin));
    if (root != NULL) {
      parse_config(root);
      sxml_delete_node(root);
      sdb = sqlite_open(SESSION_DB, 0, NULL);
    }
    fclose(fin);
  }
}

static void
parse_config(root)
	sxml_node_t *	root;
{
  sxml_node_t *	node;

  node = sxml_find_element(root, "wyvern-session-configuration", NULL, NULL);
  if (node != NULL) {
    sxml_node_t *	cp;

    for (cp = node->child; cp != NULL; cp = cp->next) {
      if (cp->type != SXML_ELEMENT) { continue; }
      if (cp->value.element.name == NULL) { continue; }

      if (ws_strcmp(cp->value.element.name, "session-information") == 0) {
	sxml_node_t *	np;
	const char *	args[3] = { NULL, NULL, NULL };

	for (np = cp->child; np != NULL; np = np->next) {
	  const char *	content;

	  if (np->type != SXML_ELEMENT) { continue; }
	  if (np->value.element.name == NULL) { continue; }

	  content = sxml_get_content(np->child);
	  if (ws_strcmp(np->value.element.name, "server-name") == 0) {
	    args[0] = content;
	  }
	  else if (ws_strcmp(np->value.element.name, "region-path") == 0) {
	    args[1] = content;
	  }
	  else if (ws_strcmp(np->value.element.name, "login-page") == 0) {
	    args[2] = content;
	  }
	  else {
	    parse_error(np->value.element.name);
	  }
	}
	if (args[0] != NULL && args[1] != NULL && args[2] != NULL) {
	  wmSession *	new;

	  new = (wmSession *)ws_malloc(sizeof(wmSession));
	  if (new != NULL) {
	    new->name = ws_strdup(args[0]);
	    new->path = ws_strdup(args[1]);
	    new->plen = strlen(args[1]);
	    new->page = ws_strdup(args[2]);
	    if (smgr != NULL) {
	      new->next = smgr;
	    }
	    else {
	      new->next = NULL;
	    }
	    smgr = new;
	  }
	}
      }
      else {
	parse_error(cp->value.element.name);
      }
    }
  }
}

static void
parse_error(tag)
	const char *	tag;
{
  wv_die("[mod_session2] error: unknown tag: %s", tag);
}

/*****************************************************************************/

/*
 * Function Name:	wm_session_finish
 * Description:		delete CGI object
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_session_finish(void)
{
  register wmSession *	sp;
  register wmSession *	sp_next;

  if (sdb != NULL) {
    sqlite_close(sdb);
    sdb = NULL;
  }

  for (sp = smgr; sp != NULL; sp = sp_next) {
    sp_next = sp->next;
    if (sp->name != NULL) { free((char *)sp->name); }
    if (sp->path != NULL) { free((char *)sp->path); }
    if (sp->page != NULL) { free((char *)sp->page); }
    free(sp);
  }
}

/*
 * Function Name:	wm_validate_session
 * Description:		validate a session
 * Arguments:		message
 * Return Value:	Successful: 0, otherwise: -1
 */
static int
wm_validate_session(req)
	register Request *	req;
{
  wmSession *	sp;
  int		ret = -1;

#if	DEBUG
  do {
    fprintf(stderr, "DEBUG[mod_session2] request: %s:%d%s\n",
	    req->vhost->serverName, req->soci.port, req->uri);
    fprintf(stderr, "DEBUG[mod_session2] query_string: %s\n",
	    req->query_string);
    fprintf(stderr, "DEBUG[mod_session2] cookie: %s\n", req->header_cookie);
  } while (0);
#endif	/* DEBUG */

  sp = lookup(req->vhost->serverName, req->uri);
  if (sp == NULL) {
    return 0;
  }
  /* session id in query-string */
  if (req->query_string != NULL) {
    char *	s = ws_strdup(req->query_string);
    if (s != NULL) {
      char *	id;

      id = get_sid_from_query(s);
      if (id != NULL) {
	int	n;
	long	expire;
	long	interval;

	free((char *)s);
	n = validate_session_id(id, &expire, &interval);
	if (n == WM_SID_VALID) {
	  return 0;
	}
      }
    }
  }
#if	DEBUG
  wv_syslog(WLL_LOG_DEBUG, "No-QueryString: %s", req->query_string);
#endif	/* DEBUG */

  /* session id in cookie */
  if (req->header_cookie != NULL) {
    char *	s = ws_strdup(req->header_cookie);
    if (s != NULL) {
      char *	id;

      id = get_sid_from_cookie(s);
#if	DEBUG
      wv_syslog(WLL_LOG_DEBUG, "Cookie id: %s", id);
#endif	/* DEBUG */
      if (id != NULL) {
	int	n;
	char *	c;	/* new cookie */
	long	expire;
	long	interval;

	free((char *)s);
	n = validate_session_id(id, &expire, &interval);
#if	DEBUG
	switch (n) {
	  case WM_SID_VALID:
	    wv_syslog(WLL_LOG_DEBUG, "Valid: %s", id);
	    break;
	  case WM_SID_EXPIRED:
	    wv_syslog(WLL_LOG_DEBUG, "Expired: %s", id);
	    break;
	  case WM_SID_INVALID:
	    wv_syslog(WLL_LOG_DEBUG, "Invalid: %s", id);
	    break;
	}
#endif	/* DEBUG */
	if (n == WM_SID_VALID) {
	  asprintf(&c, SESSION_ID "=%s; Path=%s; Max-Age=%ld; Version=1",
		   id, sp->path, interval);
	  req->header_set_cookie = c;
	  free((char *)id);
	  ret = 0;
#if	DEBUG
	  wv_syslog(WLL_LOG_DEBUG, "Set-Cookie: %s", c);
#endif	/* DEBUG */
	}
      }
    }
  }

#if	DEBUG
  else {
    wv_syslog(WLL_LOG_DEBUG, "No-Cookie: ");
  }
  wv_syslog(WLL_LOG_DEBUG, "Request: %s:%d%s",
	    req->vhost->serverName, req->soci.port, req->uri);
#endif	/* DEBUG */
  /* invalid session id */
  if (ret == -1) {
    char *	uri;

    if (req->soci.port == HTTP_PORT) {
      asprintf(&uri, "http://%s%s", req->vhost->serverName, sp->page);
    }
    else {
      asprintf(&uri, "http://%s:%d%s",
	       req->vhost->serverName, req->soci.port, sp->page);
    }
#if	DEBUG
    wv_syslog(WLL_LOG_DEBUG, "Location: %s", uri);
#endif	/* DEBUG */
    req->header_location = uri;
    req->status		 = STATUS_MOVED_PERMANENTLY;
  }

  return ret;
}

static wmSession *
lookup(name, uri)
	const char *	name;
	const char *	uri;
{
  register wmSession *	sp;

  for (sp = smgr; sp != NULL; sp = sp->next) {
    if (ws_strcmp(uri, sp->page) != 0 &&
	ws_strcmp(name, sp->name) == 0 &&
	ws_strncmp(uri, sp->path, sp->plen) == 0) {
      return sp;
    }
  }

  return NULL;
}

/*
 * Function Name:	validate_session_id
 * Description:		validate a session id
 * Arguments:		session id
 * Return Value:	Successful: 0, otherwise: -1
 */
static int
validate_session_id(id, expire, interval)
	const char *	id;
	long *		expire;
	long *		interval;
{
  struct timeval	now;

  if (gettimeofday(&now, (struct timezone *)NULL) != -1) {
    char *	sql;

    asprintf(&sql, "SELECT expire,interval FROM " SESSION_TABLE " WHERE id = '%s';", id);
    if (sql != NULL) {
      sqlite_vm *	svm;
      int		rc;

#if	DEBUG
      wv_syslog(WLL_LOG_DEBUG, "Validator: %s", sql); 
#endif	/* DEBUG */
      rc = sqlite_compile(sdb, sql, NULL, &svm, NULL);	free((char *)sql);
      if (rc == SQLITE_OK) {
	int		n;
	const char **	value;

	rc = sqlite_step(svm, &n, &value, NULL);
	if (rc == SQLITE_ROW && n == 2) {
	  *expire    = strtol(value[0], (char **)NULL, 10);
	  *interval  = strtol(value[1], (char **)NULL, 10);
	  sqlite_finalize(svm, NULL);
	  if (now.tv_sec > *expire) {
	    return WM_SID_EXPIRED;
	  }
	  else {
	    sqlite_exec_printf(sdb,
		"BEGIN;"
		"UPDATE " SESSION_TABLE " SET expire = %ld WHERE id = '%s';"
		"COMMIT;",
		NULL, NULL, NULL, now.tv_sec + *interval * 60, id);
	    return WM_SID_VALID;
	  }
	}
#if	DEBUG
	wv_syslog(WLL_LOG_DEBUG, "No exist: rc=%d id=%s", rc, id);
#endif	/* DEBUG */
      }
    }
  }

  return WM_SID_INVALID;
}

/*
 * FUNCTION NAME:	get_sid_from_query
 * DESCRIPTION:		parse CGI query data
 * ARGUMENTS:		query string
 * RETURN VALUE:	Successful: id, otherwise: NULL
 */
static char *
get_sid_from_query(query_string)
	char *	query_string;
{
  char *	string = query_string;

  while ((string = strtok(string, "&")) != NULL) {
    char *	name;
    char *	value;

    name  = string;
    value = (char *)strchr(string, '=');
    if (value != NULL) {
      *value++ = EOL;
      if (ws_strcmp(name, SESSION_ID) == 0) {
       	return value;
      }
    }
    string = NULL;	/* DON'T REMOVE THIS LINE. */
  }

  return NULL;
}

/*
 * FUNCTION NAME:	get_sid_from_cookie
 * DESCRIPTION:		parse Cookie string
 * ARGUMENTS:		cookie string
 * RETURN VALUE:	Successful: id, otherwise: NULL
 */
static char *
get_sid_from_cookie(http_cookie)
	char *	http_cookie;
{
  char *	string = http_cookie;

  while ((string = strtok(string, ";,")) != NULL) {
    char *	name;	/* parameter name */
    char *	value;	/* parameter value */

    /* skip white space */
    while (*string == SPC || *string == TAB) {
      string++;
    }

    if (parse_string(string, &name, &value) == 0) {
      if (ws_strcmp(name, SESSION_ID) == 0) {
	free((char *)name);
	return value;
      }
    }
    string = NULL;	/* DON'T REMOVE! */
  }

  return NULL;
}

/*
 * FUNCTION NAME:	parse_string
 * DESCRIPTION:		parse NAME=VALUE string of cookie
 * ARGUMENTS:		string, name pointer, value pointer
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
parse_string(string, name, value)
	register char *	string;
	char **		name;
	char **		value;
{
  register char *	p;

#if	DEBUG
  wv_syslog(WLL_LOG_DEBUG, "Cookie: %s", string);
#endif	/* DEBUG */
  /* initialize */
  *name = *value = NULL;

  /* $Secure and $Discard is special parameter without a value */
  if (ws_strcasecmp(string, "$Secure")  == 0 ||
      ws_strcasecmp(string, "$Discard") == 0) {	/* see RFC2965 */
    *name = ws_strdup(string);
    return 0;
  }

  /* NAME */
  p = strchr(string, '=');
  if (p == NULL) {
    return -1;
  }
  *p++  = EOL;
  *name = ws_strdup(string);
  /* VALUE */
  if (*p == '"') {
    char *	q;

    q = ++p;
    p = strchr(q, '"');
    if (p) {
      *p = EOL;
    }
    *value = ws_strdup(q);
  }
  else {
    *value = ws_strdup(p);
  }

  if (*name == NULL || *value == NULL) {
    wv_free((char *)*name);
    wv_free((char *)*value);
    return -1;
  }

  return 0;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_session_register_hooks(void)
{
  wm_hook_fixups(wm_validate_session, HOOK_FIXUP_REQUEST);
}

static wmModule	mod_session2 = {
  WM_MAGIC_NUMBER_MAJOR,	/* version */
  WM_UNSECURE,			/* unsecure */
  "mod_session2",		/* name */
  "Session Management Module version 1.1.0",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* depend */
  NULL,				/* handle */
  (void *)wm_session_init,		/* init */
  (void *)wm_session_finish,	/* finish */
  (void *)wm_session_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_session2;
}
