/*****************************************************************************
 *
 * FILE:	cookie.c
 * DESCRIPTION:	Cockatrice: cookie manipulator
 * DATE:	Sat, Mar 16 2002
 * UPDATED:	Sun, Aug 19 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: cookie.c,v 1.3 2007/08/19 16:48:36 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "cockatrice.h"
#include "module.h"
#include "functions.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	COOKIE_TABLE_SIZE	(11)
typedef struct _cookie2 {	/* cookie structure (cf. RFC2965) */
  char *		name;		/* NAME */
  char *		value;		/* VALUE */
  char *		comment;	/* $Comment */
  char *		commentURL;	/* $CommentURL */
  bool			discard;	/* $Discard */
  char *		domain;		/* $Domain */
  unsigned int		max_age;	/* $Max-Age */
  char *		path;		/* $Path */
  char *		port;		/* $Port */
  bool			secure;		/* $Secure */
  unsigned int		version;	/* $Version=1 */
  bool			use_cookie2;	/* use Set-Cookie2? */
  struct _cookie2 *	next;
} Cookie;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static char *	cm_cookie_get(const char *);
static int	cm_cookie_set_all(const char *, const char *, const char *,
				  const char *, bool, const char *,
				  unsigned int, const char *, const char *,
				  bool, unsigned int);
static int	cm_cookie_set_name(const char *, const char *);
static int	cm_cookie_set_comment(const char *, const char *);
static int	cm_cookie_set_commentURL(const char *, const char *);
static int	cm_cookie_set_discard(const char *, bool);
static int	cm_cookie_set_domain(const char *, const char *);
static int	cm_cookie_set_max_age(const char *, unsigned int);
static int	cm_cookie_set_path(const char *, const char *);
static int	cm_cookie_set_port(const char *, const char *);
static int	cm_cookie_set_secure(const char *, bool);
static void	cm_cookie_send_header(const char *);
static void	cm_cookie_redirect(const char *);

static void	cleanup(Cookie *);
static int	parse_cookie(void);
static int	parse_string(register char *, char * *, char * *);
static Cookie *		lookup(const char *);
static unsigned int	hash(const char *);
static void	send_cookie_header(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static Cookie *	cookie[COOKIE_TABLE_SIZE];	/* Cookie field */
static Cookie *	setCookie[COOKIE_TABLE_SIZE];	/* Set-Cookie field */

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	cc_cookie_init
 * DESCRIPTION:		initialize a cookie table
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
void
cc_cookie_init(void)
{
  register int	i;

  for (i = 0; i < COOKIE_TABLE_SIZE; i++) {
    cookie[i] = setCookie[i] = NULL;
  }
  parse_cookie();
}

/*
 * FUNCTION NAME:	cc_cookie_finish
 * DESCRIPTION:		release memory area of cookie table
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
void
cc_cookie_finish(void)
{
  unsigned int	i;

  for (i = 0; i < COOKIE_TABLE_SIZE; i++) {
    if (cookie[i] != NULL) {
      cleanup(cookie[i]);
      cookie[i] = NULL;
    }
    if (setCookie[i] != NULL) {
      cleanup(setCookie[i]);
      setCookie[i] = NULL;
    }
  }
}

static void
cleanup(c)
	Cookie *	c;
{
  register Cookie *	cp;
  register Cookie *	cp_next;

  for (cp = c; cp != NULL; cp = cp_next) {
    cp_next = cp->next;
    cc_free(cp->name);
    cc_free(cp->value);
    cc_free(cp->comment);
    cc_free(cp->commentURL);
    cc_free(cp->domain);
    cc_free(cp->path);
    cc_free(cp->port);
    free(cp);
  }
}

/*
 * FUNCTION NAME:	parse_cookie
 * DESCRIPTION:		parse Cookie string
 * ARGUMENTS:		none
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
parse_cookie(void)
{
  char *	string;

  string = getenv("HTTP_COOKIE");
  if (string == NULL) {
    string = getenv("HTTP_COOKIE2");
    if (string == NULL) {
      return -1;	/* No cookies? or error? */
    }
  }

  while ((string = strtok(string, ";,")) != NULL) {
    Cookie	c;	/* cookie */
    char *	Pname;	/* parameter name */
    char *	Pvalue;	/* parameter value */
    char *	Tname;	/* temporary pointer to call free() */
    char *	Tvalue;	/* temporary pointer to call free() */

    /* skip white space */
    while (*string == SPC || *string == TAB) {
      string++;
    }

    memset((char *)&c, 0, sizeof(c));
    if (parse_string(string, &Pname, &Pvalue) == 0) {
      c.name  = Tname  = cc_uri_unescape(Pname);	free(Pname);
      c.value = Tvalue = cc_uri_unescape(Pvalue);	free(Pvalue);

      /* reserved parameter */
      if (c.name[0] == '$') {
	if (strcasecmp(c.name, "$Version") == 0) {
	  c.version = (unsigned int)strtol(c.value, (char **)NULL, 10);
	}
	else if (strcasecmp(c.name, "$Domain") == 0) {
	  c.domain = c.value;
	}
	else if (strcasecmp(c.name, "$Path") == 0) {
	  c.path = c.value;
	}
	else if (strcasecmp(c.name, "$Max-Age") == 0) {
	  c.max_age = (unsigned int)strtol(c.value, (char **)NULL, 10);
	}
	else if (strcasecmp(c.name, "$Secure") == 0) {
	  c.secure = true;
	}
	else if (strcasecmp(c.name, "$Comment") == 0) {
	  c.comment = c.value;
	}
	else if (strcasecmp(c.name, "$CommentURL") == 0) {	/* RFC2965 */
	  c.commentURL = c.value;
	}
	else if (strcasecmp(c.name, "$Discard") == 0) {	/* RFC2965 */
	  c.discard = true;
	}
	else if (strcasecmp(c.name, "$Port") == 0) {	/* RFC2965 */
	  c.port = c.value;
	}
	else {
	  /* invalid reserved parameter */
	}
      }
      /* user defined parameter */
      else {
#if	ENABLE_ICONV || ENABLE_WKF
	char *		Kname;
	char *		Kvalue;
	unsigned int	len;

	len   = strlen(c.name) + 1;
	Kname = (char *)cc_calloc(len, sizeof(char));
	if (Kname != NULL && cc_conv_any2euc(c.name, len, Kname, len) == 0) {
  	  c.name = Kname;
   	}
	len    = strlen(c.value) + 1;
	Kvalue = (char *)cc_calloc(len, sizeof(char));
	if (Kvalue != NULL && cc_conv_any2euc(c.value, len, Kvalue, len) == 0) {
	  c.value = Kvalue;
	}
#endif	/* ENABLE_WKF */
	cm_cookie_set_all(c.name, c.value, c.comment, c.commentURL, c.discard,
			  c.domain, c.max_age, c.path, c.port, c.secure,
			  c.version);
#if	ENABLE_ICONV || ENABLE_WKF
	cc_free(Kname);
	cc_free(Kvalue);
#endif	/* ENABLE_WKF */
      }
      free(Tname);
      free(Tvalue);
    }
    string = NULL;	/* DON'T REMOVE! */
  }

  return 0;
}

/*
 * FUNCTION NAME:	parse_string
 * DESCRIPTION:		parse NAME=VALUE string of cookie
 * ARGUMENTS:		string, name pointer, value pointer
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
parse_string(string, name, value)
	register char *	string;
	char **		name;
	char **		value;
{
  register char *	p;

  /* initialize */
  *name = *value = NULL;

  /* $Secure and $Discard is special parameter without a value */
  if (strcasecmp(string, "$Secure")  == 0 ||
      strcasecmp(string, "$Discard") == 0) {	/* see RFC2965 */
    *name = cc_strdup(string);
    return 0;
  }

  /* NAME */
  p = strchr(string, '=');
  if (p == NULL) {
    return -1;
  }
  *p++  = EOL;
  *name = cc_strdup(string);
  /* VALUE */
  if (*p == '"') {
    char *	q;

    q = ++p;
    p = strchr(q, '"');
    if (p) {
      *p = EOL;
    }
    *value = cc_strdup(q);
  }
  else {
    *value = cc_strdup(p);
  }

  if (*name == NULL || *value == NULL) {
    cc_free(*name); 
    cc_free(*value);
    return -1;
  }

  return 0;
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	cm_cookie_get
 * DESCRIPTION:		get Cookie value from the table
 * ARGUMENTS:		name
 * RETURN VALUE:	value
 */
static char *
cm_cookie_get(name)
	const char *	name;
{
  if (name != NULL) {
    unsigned int	h;	/* hash value */
    register Cookie *	cp;	/* working pointer */

    h = hash(name);
    for (cp = cookie[h]; cp != NULL; cp = cp->next) {
      if (cp->name != NULL && strcmp(name, cp->name) == 0) {
	return cp->value;
      }
    }
  }

  return NULL;
}

/*
 * FUNCTION NAME:	cm_cookie_set_all
 * DESCRIPTION:		store 'NAME=VALUE' of cookie into the table
 * ARGUMENTS:		name, value, comment, commentURL, discard, domain,
 *			max_age, path, port, secure, version
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_cookie_set_all(name, value, comment, commentURL, discard, domain, max_age,
		  path, port, secure, version)
	const char *	name;
	const char *	value;
	const char *	comment;
	const char *	commentURL;
	bool		discard;
	const char *	domain;
	unsigned int	max_age;
	const char *	path;
	const char *	port;
	bool		secure;
	unsigned int	version;
{
  Cookie *	cp;	/* new parameter pointer */
  unsigned int	h;	/* hash value */

  /* reject no name / no value */
  if (name == NULL || value == NULL) {
    return -1;
  }

  cp = (Cookie *)cc_calloc(1, sizeof(Cookie));
  if (cp == NULL) {
    return -1;
  }

  h		= hash(name);
  cp->name	= cc_strdup(name);
  cp->value	= cc_strdup(value);
  cp->comment	= comment    ? cc_strdup(comment)    : NULL;
  cp->commentURL= commentURL ? cc_strdup(commentURL) : NULL;
  cp->discard	= discard;
  cp->domain	= domain ? cc_strdup(domain) : NULL;
  cp->max_age	= max_age;
  cp->path	= path ? cc_strdup(path) : NULL;
  cp->port	= port ? cc_strdup(port) : NULL;
  cp->secure	= secure;
  cp->version	= version;
  cp->next	= cookie[h];
  cookie[h]	= cp;

  return 0;
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	cm_cookie_set_name
 * DESCRIPTION:		set Cookie Name and Value
 * ARGUMENTS:		name, value
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_cookie_set_name(name, value)
	const char *	name;
	const char *	value;
{
  Cookie *	cp;	/* new parameter pointer */
  unsigned int	h;	/* hash value */

  /* reject no name / no value */
  if (name == NULL || value == NULL) {
    cgi_errno = CGIERR_INVALID_ARGUMENTS;
    return -1;
  }
  cp = (Cookie *)cc_calloc(1, sizeof(Cookie));
  if (cp == NULL) {
    return -1;
  }

  h		= hash(name);
  cp->name	= cc_strdup(name);
  cp->value	= cc_strdup(value);
  cp->comment	= NULL;
  cp->commentURL= NULL;
  cp->discard	= false;
  cp->domain	= NULL;
  cp->max_age	= 0;
  cp->path	= NULL;
  cp->port	= NULL;
  cp->secure	= false;
  cp->version	= 1;
  cp->next	= setCookie[h];
  setCookie[h]	= cp;

  return 0;
}

/*
 * FUNCTION NAME:	cm_cookie_set_comment
 * DESCRIPTION:		set Cookie Comment
 * ARGUMENTS:		name, comment
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_cookie_set_comment(name, comment)
	const char *	name;
	const char *	comment;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL && comment != NULL) {
    cp->comment = cc_strdup(comment);
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cm_cookie_set_commentURL
 * DESCRIPTION:		set Cookie CommentURL
 * ARGUMENTS:		name, url
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_cookie_set_commentURL(name, url)
	const char *	name;
	const char *	url;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL && url != NULL) {
    cp->commentURL  = cc_strdup(url);
    cp->use_cookie2 = true;
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cm_cookie_set_discard
 * DESCRIPTION:		set Cookie Discard
 * ARGUMENTS:		name, discard
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_cookie_set_discard(name, discard)
	const char *	name;
	bool		discard;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL) {
    cp->discard	    = discard;
    cp->use_cookie2 = true;
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cm_cookie_set_domain
 * DESCRIPTION:		set Cookie Domain
 * ARGUMENTS:		name, domain
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_cookie_set_domain(name, domain)
	const char *	name;
	const char *	domain;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL && domain != NULL) {
    cp->domain = cc_strdup(domain);
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cm_cookie_set_max_age
 * DESCRIPTION:		set Cookie MaxAge
 * ARGUMENTS:		name, max-age
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_cookie_set_max_age(name, max_age)
	const char *	name;
	unsigned int	max_age;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL) {
    cp->max_age = max_age;
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cm_cookie_set_path
 * DESCRIPTION:		set Cookie Path
 * ARGUMENTS:		name, path
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_cookie_set_path(name, path)
	const char *	name;
	const char *	path;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL && path != NULL) {
    cp->path = cc_strdup(path);
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cm_cookie_set_port
 * DESCRIPTION:		set Cookie Port
 * ARGUMENTS:		name, portlist
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_cookie_set_port(name, portlist)
	const char *	name;
	const char *	portlist;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL && portlist != NULL) {
    cp->port	    = cc_strdup(portlist);
    cp->use_cookie2 = true;
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cm_cookie_set_secure
 * DESCRIPTION:		set Cookie Secure
 * ARGUMENTS:		name, secure
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_cookie_set_secure(name, secure)
	const char *	name;
	bool		secure;
{
  Cookie *	cp = lookup(name);

  if (cp != NULL) {
    cp->secure = secure;
    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	lookup
 * DESCRIPTION:		look up a SetCookie data for Cookie name
 * ARGUMENTS:		name
 * RETURN VALUE:	Successful: Cookie pointer, otherwise: NULL
 */
static Cookie *
lookup(name)
	const char *	name;
{
  if (name != NULL) {
    unsigned int	h;	/* hash value */
    register Cookie *	cp;	/* working pointer */

    h = hash(name);
    for (cp = setCookie[h]; cp != NULL; cp = cp->next) {
      if (cp->name != NULL && strcmp(name, cp->name) == 0) {
	return cp;
      }
    }
  }

  return NULL;
}

/*
 * FUNCTION NAME:	hash
 * DESCRIPTION:		calculates a hash value of string
 * ARGUMENTS:		string
 * RETURN VALUE:	hash value
 */
static unsigned int
hash(str)
	const char *	str;
{
  register char *	p;
  register unsigned int	h;

  p = (char *)str;
  for (h = *p++; *p != EOL; p++) {
    h += *p;
  }

  return h % COOKIE_TABLE_SIZE;
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	send_cookie_header
 * DESCRIPTION:		send Cookie header field
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
send_cookie_header(void)
{
  register int	i;

  /* Set-Cookie response field */
  for (i = 0; i < COOKIE_TABLE_SIZE; i++) {
    register Cookie *	cp;

    if (setCookie[i] == NULL) {
      continue;
    }
    for (cp = setCookie[i]; cp != NULL; cp = cp->next) {
      fputs("Set-Cookie", stdout);
      if (cp->use_cookie2) {
	fputs("2", stdout);
      }
      fprintf(stdout, ": %s=%s",
	      cc_uri_escape(cp->name), cc_uri_escape(cp->value));
      if (cp->comment) {
	fprintf(stdout, "; Comment=%s", cp->comment);
      }
      if (cp->commentURL) {
	fprintf(stdout, "; CommentURL=\"%s\"", cp->commentURL);
      }
      if (cp->discard) {
	fputs("; Discard", stdout);
      }
      if (cp->domain) {
	fprintf(stdout, "; Domain=%s", cp->domain);
      }
      if (cp->max_age > 0) {
	fprintf(stdout, "; Max-Age=%d", cp->max_age);
      }
      if (cp->path) {
	fprintf(stdout, "; Path=%s", cp->path);
      }
      if (cp->port) {
	fprintf(stdout, "; Port=\"%s\"", cp->port);
      }
      if (cp->secure) {
	fputs("; Secure", stdout);
      }
      fprintf(stdout, "; Version=%d" CRLF, cp->version);
    }
  }
}

/*
 * FUNCTION NAME:	cm_cookie_send_header
 * DESCRIPTION:		send response header with Cookie
 * ARGUMENTS:		Content-Type
 * RETURN VALUE:	none
 */
static void
cm_cookie_send_header(contentType)
	const char *	contentType;
{
  /* Content-Type response field */
  fputs("Content-Type: ", stdout);
  if (contentType != NULL) {
    fprintf(stdout, "%s" CRLF, contentType);
  }
  else {
    fputs("text/html" CRLF, stdout);
  }
  send_cookie_header();
  fputs(CRLF, stdout);
}

/*
 * FUNCTION NAME:	cm_cookie_redirect
 * DESCRIPTION:		flush out to Location with Cookie header
 * ARGUMENTS:		URI
 * RETURN VALUE:	none
 */
static void
cm_cookie_redirect(uri)
	const char *	uri;
{
  fputs("Status: 301 Moved" CRLF, stdout);
  fprintf(stdout, "Location: %s" CRLF, uri);
  send_cookie_header();
  fputs(CRLF, stdout);
}

/*****************************************************************************/

static cmCookie	cookie_funcs = {
  cm_cookie_get,
  cm_cookie_set_all,
  cm_cookie_set_name,
  cm_cookie_set_comment,
  cm_cookie_set_commentURL,
  cm_cookie_set_discard,
  cm_cookie_set_domain,
  cm_cookie_set_max_age,
  cm_cookie_set_path,
  cm_cookie_set_port,
  cm_cookie_set_secure
};

void
cc_cookie_register_hooks(void)
{
  cc_hook_response_header(cm_cookie_send_header);
  cc_hook_redirect_uri(cm_cookie_redirect);
  cc_hook_module_cookie(&cookie_funcs);
}
