/*****************************************************************************
 *
 * FILE:	mod_session.c
 * DESCRIPTION:	Cockatrice: session manager with SQLite3 (module)
 * DATE:	Mon, Jul  5 2004
 * UPDATED:	Mon, Aug 20 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2004-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2004-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_session.c,v 1.1 2007/08/19 16:49:15 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif	/* HAVE_SYS_TIME_H */
#include <limits.h>
#include <errno.h>
#include <sha.h>
#include <sqlite3.h>
#include "cockatrice.h"
#include "module.h"
#include "ccpath.h"
#include "functions.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	DEF_SESSION_EXPIRES	(20)	/* 20 min. */

typedef struct {
  sqlite3 *	conn;
  char *	name;	/* SCRIPT_NAME */
  char *	addr;	/* REMOTE_ADDR */
  char *	port;	/* REMOTE_PORT */
} csession;

#if	HAVE_DEV_URANDOM
#define	DEV_URANDOM	"/dev/urandom"
#endif	/* HAVE_DEV_URANDOM */

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	cm_session_init(void);
static void	cm_session_finish(void);
static char *	cm_session_open(const char * sname, long expiry);
static int	cm_session_close(const char * sname, const char * id);
static int	cm_session_validate_id(const char * sname, const char * id);
static int	cm_session_delete_id(const char * sname, const char * id);
static long	cm_session_get_expire(const char * sname, const char * id);
static int	get_limits(const char * sname, const char * id,
			   long * expire, long * interval);
static long	urandom(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static csession	cs = { NULL, NULL, NULL, NULL };

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

static void
cm_session_init(void)
{
  if (sqlite3_open(SESSION_DB, &cs.conn) != SQLITE_OK) {
    fprintf(stderr, "Failed to open database: %s\n", sqlite3_errmsg(cs.conn));
    exit(1);
  }
  cs.name = cc_strdup(getenv("SCRIPT_NAME"));
  cs.addr = cc_strdup(getenv("REMOTE_ADDR"));
  cs.port = cc_strdup(getenv("REMOTE_PORT"));
}

static void
cm_session_finish(void)
{
  if (cs.conn)	 { sqlite3_close(cs.conn); }
  if (cs.name)	 { free(cs.name); }
  if (cs.addr)	 { free(cs.addr); }
  if (cs.port)	 { free(cs.port); }
}

/*
 * FUNCTION NAME:	cm_session_open
 * DESCRIPTION:		open session database
 * ARGUMENTS:		session name, expiry
 * RETURN VALUE:	Successful: session id, otherwise: NULL
 */
static char *
cm_session_open(sname, expiry)
	const char *	sname;
	long		expiry;	/* minutes */
{
  struct timeval	now;

  if (gettimeofday(&now, (struct timezone *)NULL) == 0) {
    char *	text;

    asprintf(&text, "%ld%s%u%s%s%s%ld",
	   now.tv_sec, cs.name, getpid(), cs.addr, cs.port, sname, urandom());
    if (text) {
      SHA_CTX		ctx;
      unsigned char	id[SHA_DIGEST_LENGTH * 2 + 1];
      char *		sql;

      memset(id, 0, sizeof(id));
      SHA1_Init(&ctx);
      SHA1_Update(&ctx, text, strlen(text));
      SHA1_End(&ctx, id);

      if (expiry < 0) {
	expiry = DEF_SESSION_EXPIRES;
      }
      expiry *= 60;	/* to seconds */

      sql = sqlite3_mprintf("INSERT INTO csession"
			    " VALUES('%q','%q','%ld','%ld');",
			    sname, id, now.tv_sec + expiry, expiry);
      if (sql != NULL) {
	int	rc;

	rc = sqlite3_exec(cs.conn, sql, NULL, NULL, NULL);
	sqlite3_free(sql);
	if (rc == SQLITE_OK) {
#if	DEBUG
	  fprintf(stderr, "DEBUG[session_open] id=%s\n", id);
#endif	/* DEBUG */
	  return cc_strdup(id);
	}
      }
    }
  }

  cgi_errno = CGIERR_SESSION_OPEN_FAILURE;

  return NULL;
}

static int
cm_session_close(sname, id)
	const char *	sname;
	const char *	id;
{
  struct timeval	now;
  int			status = -1;

  if (gettimeofday(&now, (struct timezone *)NULL) == 0) {
    char *	sql;

    sql = sqlite3_mprintf("DELETE FROM csession WHERE name='%q' AND id='%q';",
			  sname, id);
    if (sql != NULL) {
      int	rc;

      rc = sqlite3_exec(cs.conn, sql, NULL, NULL, NULL);
      sqlite3_free(sql);
      status = (rc == SQLITE_OK) ? 0 : -1;
    }
    sql = sqlite3_mprintf("DELETE FROM csession WHERE expire < %ld;",
			  now.tv_sec);
    if (sql != NULL) {
      sqlite3_exec(cs.conn, sql, NULL, NULL, NULL);
      sqlite3_free(sql);
    }
  }

  return status;
}

/*
 * FUNCTION NAME:	cm_session_validate_id
 * DESCRIPTION:		validate a session id
 * ARGUMENTS:		session name, id
 * RETURN VALUE:	Valid: 0, Expired: 1, otherwise: -1
 */
static int
cm_session_validate_id(sname, id)
	const char *	sname;
	const char *	id;
{
  struct timeval	now;

  if (gettimeofday(&now, (struct timezone *)NULL) == 0) {
    long	expire;		/* expired timestamp */
    long	interval;	/* interval to update expire */
    char *	sql;

    if (get_limits(sname, id, &expire, &interval) == -1) {
      cgi_errno = CGIERR_INTERNAL_LIBRARY_ERROR;
      return -1;
    }

    if (cgi_errno != CGIERR_NO_ERROR) {
      cgi_errno = CGIERR_NO_SESSION_ID;
      return CC_SID_INVALID;
    }

#if	DEBUG
    fprintf(stderr, "DEBUG[validate_id] expire=%ld interval=%ld\n",
	    expire, interval);
#endif	/* DEBUG */
    if (now.tv_sec > expire) {
      int	rc;

      cgi_errno	= CGIERR_EXPIRED_SESSION_ID;
      rc	= cm_session_delete_id(sname, id);

      return (rc == 0) ? CC_SID_EXPIRED : CC_SID_INVALID;
    }

    sql = sqlite3_mprintf("UPDATE csession SET expire=%ld"
			  " WHERE name='%q' AND id='%q';",
			  now.tv_sec + interval, sname, id);
    if (sql != NULL) {
      int	rc;

      rc = sqlite3_exec(cs.conn, sql, NULL, NULL, NULL);
      sqlite3_free(sql);
      if (rc == SQLITE_OK) {
	return CC_SID_VALID;
      }
    }
    cgi_errno = CGIERR_SESSION_ID_UPDATE_FAILURE;
  }

  return CC_SID_INVALID;
}

/*
 * FUNCTION NAME:	cm_session_delete_id
 * DESCRIPTION:		delete a session id
 * ARGUMENTS:		session name, id
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_session_delete_id(sname, id)
	const char *	sname;
	const char *	id;
{
  char *	sql;

  sql = sqlite3_mprintf("DELETE FROM csession WHERE name='%q' AND id='%q';",
			sname, id);
  if (sql != NULL) {
    int	rc;

    rc = sqlite3_exec(cs.conn, sql, NULL, NULL, NULL);
    sqlite3_free(sql);
    if (rc == SQLITE_OK) {
      return 0;
    }
  }
  cgi_errno = CGIERR_SESSION_ID_DELETE_FAILURE;

  return -1;
}

/*
 * FUNCTION NAME:	cm_session_get_expire
 * DESCRIPTION:		get a session expiry
 * ARGUMENTS:		session name, id
 * RETURN VALUE:	successful: expire, otherwise: -1
 */
static long
cm_session_get_expire(sname, id)
	const char *	sname;
	const char *	id;
{
  long	expire;
  long	interval;

  if (get_limits(sname, id, &expire, &interval) == -1 ||
      cgi_errno == CGIERR_NO_SESSION_ID) {
    return -1;
  }

  return expire;
}

static int
get_limits(sname, id, expire, interval)
	const char *	sname;
	const char *	id;
	long *		expire;		/* set returned value */
	long *		interval;	/* set returned value */
{
  char *	sql;
  int		status = -1;

  sql = sqlite3_mprintf("SELECT expire,interval FROM csession"
			" WHERE name='%q' AND id='%q';",
			sname, id);
  if (sql != NULL) {
    sqlite3_stmt *	res;
    const char *	tail;
    int			rc;

    rc = sqlite3_prepare_v2(cs.conn, sql, (int)strlen(sql), &res, &tail);
    if (rc == SQLITE_OK) {
      switch (sqlite3_step(res)) {
	case SQLITE_ROW:
	  if (sqlite3_column_count(res) == 2) {
	    *expire   = sqlite3_column_int(res, 0);
	    *interval = sqlite3_column_int(res, 1);
	  }
	  status = 0;
	  break;
	case SQLITE_DONE:
	  cgi_errno = CGIERR_NO_SESSION_ID;
	  status    = 0;
	  break;
	case SQLITE_BUSY:
	case SQLITE_MISUSE:
	case SQLITE_ERROR:
	default:
	  break;
      }
      sqlite3_finalize(res);
    }
    sqlite3_free(sql);
  }
#if	DEBUG
  fprintf(stderr, "DEBUG[get_limit] status=%d expire=%ld interval=%ld\n",
	  status, *expire, *interval);
#endif	/* DEBUG */

  return status;
}

static long
urandom(void)
{
#if	HAVE_DEV_URANDOM
  int	fd;

  fd = open(DEV_URANDOM, O_RDONLY, 0400);
  if (fd != -1) {
    long	rnd;

    (void)read(fd, &rnd, sizeof(long));
    close(fd);

    return rnd;
  }
#endif	/* HAVE_DEV_URANDOM */

  return (long)random();
}

/*****************************************************************************/

static const cmSession	session_funcs = {
  cm_session_open,
  cm_session_close,
  cm_session_validate_id,
  cm_session_delete_id,
  cm_session_get_expire
};

static void
cm_session_register_hooks(void)
{
  cc_hook_module_session(&session_funcs);
}

static ccModule	mod_session = {
  CC_MODULE_MAGIC_NUMBER_MAJOR,		/* API version */
  "mod_session",			/* module name */
  "Cockatrice: session module 1.2.0",	/* description such as version etc.. */
  NULL,					/* module handle */
  (void *)cm_session_init,		/* called when it's loaded */
  (void *)cm_session_finish,		/* called when it's unloaded */
  (void *)cm_session_register_hooks
};

ccModule *
ccModuleInfo(void)
{
  return &mod_session;
}
